from typing import Any, Optional
import base64
import pandas as pd
import pprint
import dateutil.parser
from datetime import datetime
import re


class YandexQueryResults:
    """Holds and formats query execution results"""

    Optional_Regexp = re.compile("Optional<(.*)>")

    def __init__(self, results: list[list[Any]] | list[Any]):
        self._raw_results = results
        self._results = None

    @staticmethod
    def _convert(results: list[list[Any]] | list[Any]):
        return [YandexQueryResults._convert_single(result)
                for result in results]

    @staticmethod
    def _convert_from_base64(value: str) -> str:
        return base64.b64decode(value).decode('utf-8')

    @staticmethod
    def _convert_from_datetime(value: str) -> str:
        return dateutil.parser.isoparse(value)

    @staticmethod
    def _convert_from_date(value: str) -> datetime:
        return datetime.strptime(value, "%Y-%m-%d")

    @staticmethod
    def _extract_from_optional(type):
        "Checks if type is optional i.e. Optional<Uint16>"
        matched = YandexQueryResults.Optional_Regexp.search(type)
        if matched:
            return matched.group(1)

    @staticmethod
    def _convert_from_optional(value: list[Any]) -> Optional[Any]:

        # Optional types are encoded as [[]] objects
        # If type is Uint16, value is encoded as {"rows":[[value]]}
        # If type is Optional<Uint16>, value is encoded as {"rows":[[[value]]]}
        # If value is None than result is {"rows":[[[]]]}
        # So check if len equals 1 it means that it contains value
        # if len is 0 it means it has no value i.e. value is None
        if len(value) == 1:
            return value[0]
        elif len(value) == 0:
            return None
        else:
            print(value)
            assert False

    @staticmethod
    def _get_converter(column_type):
        """Returns converter based on column type"""

        converter = None

        if column_type == "String":
            converter = YandexQueryResults._convert_from_base64
            column_type = type(str)
        elif column_type == "Datetime":
            converter = YandexQueryResults._convert_from_datetime
            column_type = type(datetime)
        elif column_type == "Date":
            converter = YandexQueryResults._convert_from_date
            column_type = type(datetime)
        elif column_type == "Timestamp":
            converter = YandexQueryResults._convert_from_datetime
            column_type = type(datetime)
        elif column_type.startswith("Optional<"):
            # If type is Optional than get base type
            converter_, column_type = YandexQueryResults._get_converter(
                YandexQueryResults._extract_from_optional(column_type))
            if converter_ is None:
                converter = YandexQueryResults._convert_from_optional
            else:
                # Remove "Optional" encoding
                # and convert resulting value as others
                def convert_from_optional(x):
                    return converter_(
                        YandexQueryResults._convert_from_optional(x))

                converter = convert_from_optional

        return (converter, column_type)

    @staticmethod
    def _convert_single(results: list[Any]) -> Any:
        converters = []
        new_column_types = []
        for column in results["columns"]:
            column_name = column["name"]
            column_type = column["type"]

            converter, column_type = YandexQueryResults._get_converter(column_type)  # noqa

            converters.append(converter)
            new_column_types.append({"name": column_name, "type": column_type})

        converted_results = []
        for row in results["rows"]:
            new_row = []
            for index, value in enumerate(row):
                converter = converters[index]
                new_row.append(
                    value if converter is None else converter(value))

            converted_results.append(new_row)

        return {"rows": converted_results, "columns": new_column_types}

    def _repr_pretty_(self, p, cycle):
        p.text(pprint.pformat(self._results))

    @property
    def results(self):
        if self._results is None:
            self._results = YandexQueryResults._convert(self.raw_results)

        return self._results

    @property
    def raw_results(self):
        return self._raw_results

    @staticmethod
    def _to_dataframe(result_set):
        return pd.DataFrame(result_set["rows"], columns=[
            column["name"] for column in result_set["columns"]])

    def to_table(self, index: Optional[int] = 0):
        if index is not None:
            return self.results[index]["rows"]
        else:
            raise ValueError("Object has multiple result sets, "
                             "please specify result set index")

    def to_dataframe(self, index: Optional[int] = 0):
        if index is not None:
            return YandexQueryResults._to_dataframe(self.results[index])
        else:
            raise ValueError("Object has multiple result sets, "
                             "please specify result set index")



from yandex_query_magic import JinjaTemplate
import pandas as pd


def test_simple_render():
    sql = "select * from {{var}}"

    rendered = JinjaTemplate.apply_template(sql, {"var": 1})
    assert rendered == "select * from 1"


def test_render_dict_as_yql():
    sql = "select * from {{dct|to_yq(name='dct')}}"

    rendered = JinjaTemplate.apply_template(sql, {"dct": {"a": "1", "b": "2"}})
    assert rendered == 'select * from ToDict(AsList(asTuple("a", "1"),asTuple("b", "2")))'


def test_render_list_as_yql():
    sql = "select * from {{lst|to_yq(name='lst')}}"

    rendered = JinjaTemplate.apply_template(sql, {"lst": [1, 2, 3]})
    assert rendered == 'select * from AsList(1l,2l,3l)'


def test_render_list_as_yql_noname():
    sql = "select * from {{lst|to_yq()}}"

    rendered = JinjaTemplate.apply_template(sql, {"lst": [1, 2, 3]})
    assert rendered == 'select * from AsList(1l,2l,3l)'


def test_render_df_as_yql():
    sql = "select * from {{df|to_yq(name='df')}}"

    df = pd.DataFrame({'_float': [1.0],
                       '_int': [1],
                       '_datetime': [pd.Timestamp('20180310')],
                       '_string': ['foo']})

    rendered = JinjaTemplate.apply_template(sql, {"df": df})
    assert rendered == 'select * from AS_TABLE(AsList(AsStruct(1.0 as `_float`,1l as `_int`,DateTime::MakeTimestamp(DateTime::Parse("%Y-%m-%d %H:%M:%S")("2018-03-10 00:00:00.000000")) as `_datetime`,"foo" as `_string`))) as `df`'  # noqa


def test_render_dict_as_yql_jinja():
    sql = """
    {% for item in items %}
        select {{item|to_yq(name='dct'+loop.index|string)}};
    {% endfor %}
        """

    rendered = JinjaTemplate.apply_template(sql, {
        "items":
            [
                {"a": 1, "b": 2},
                {"b": 2, "c": 3}
            ]
    }).strip()
    required = """select ToDict(AsList(asTuple("a", 1l),asTuple("b", 2l)));\n    \n        select ToDict(AsList(asTuple("b", 2l),asTuple("c", 3l)));"""
    assert rendered == required  # noqa



import pandas as pd
from datetime import datetime, timezone, timedelta
from yandex_query_magic import YandexQueryResults


def test_int32():
    data = [{'columns': [{'name': 'column0', 'type': 'Int32'}], 'rows': [[1]]}]
    parsed = YandexQueryResults(data).to_dataframe()
    assert parsed.equals(pd.DataFrame(columns=["column0"], data=[1]))


def test_int16_optional():
    data = [{'rows': [[[1]]], 'columns': [{'name': 'column0',
                                          'type': 'Optional<Uint16>'}]}]
    parsed = YandexQueryResults(data)
    assert parsed.to_dataframe().equals(pd.DataFrame(columns=["column0"], data=[1]))  # noqa


def test_uint32_optional():
    data = [{'rows': [[[1]]], 'columns': [{'name': 'column0',
                                          'type': 'Optional<Uint32>'}]}]
    parsed = YandexQueryResults(data).to_dataframe()
    assert parsed.equals(pd.DataFrame(columns=["column0"], data=[1]))


def test_uint32_none():
    data = [{'rows': [[[]]], 'columns': [{'name': 'column0',
                                         'type': 'Optional<Uint32>'}]}]
    parsed = YandexQueryResults(data).to_dataframe()
    assert parsed.equals(pd.DataFrame(columns=["column0"], data=[None]))


def test_utf8_optional():
    data = [{'rows': [[['a']]], 'columns': [{'name': 'column0',
                                             'type': 'Optional<Utf8>'}]}]
    parsed = YandexQueryResults(data).to_dataframe()
    assert parsed.equals(pd.DataFrame(columns=["column0"], data=['a']))


def test_date_optional():
    data = [{'rows': [[['1970-01-02']]], 'columns': [{'name': 'column0',
                                                      'type': 'Optional<Date>'}]}]  # noqa

    parsed = YandexQueryResults(data).to_dataframe()
    assert parsed.equals(pd.DataFrame(columns=["column0"],
                                      data=[datetime(1970, 1, 2)]))


def test_date():
    data = [{'rows': [['2020-01-01']], 'columns': [{'name': 'column0',
                                                    'type': 'Date'}]}]

    parsed = YandexQueryResults(data).to_dataframe()
    assert parsed.equals(pd.DataFrame(columns=["column0"],
                                      data=[datetime(2020, 1, 1)]))


def test_timestamp_optional():
    data = [{'rows': [[['2019-09-16T00:00:00Z']]], 'columns': [
        {'name': 'column0',
         'type': 'Optional<Timestamp>'}]}]

    parsed = YandexQueryResults(data).to_dataframe()
    assert parsed.equals(pd.DataFrame(columns=["column0"],
                                      data=[datetime(2019, 9, 16,
                                                     tzinfo=timezone(
                                                         offset=timedelta(seconds=0)))]))  # noqa


def test_timestamp_microseconds():
    data = [{'rows': [['2024-02-08T19:14:27.986418Z']], 'columns': [
        {'name': 'column0',
         'type': 'Timestamp'}]}]

    parsed = YandexQueryResults(data).to_dataframe()
    assert parsed.equals(pd.DataFrame(columns=["column0"],
                                      data=[datetime(year=2024,
                                                     month=2,
                                                     day=8,
                                                     hour=19,
                                                     minute=14,
                                                     second=27,
                                                     microsecond=986418,
                                                     tzinfo=timezone(
                                                         offset=timedelta(seconds=0)))]))  # noqa



from .main import YandexQuery, YandexQueryException  # noqa
from .query_results import YandexQueryResults  # noqa
from .magics import *  # noqa
from .sqltext_parser import SqlParser  # noqa
from .jinja_template import JinjaTemplate  # noqa



from __future__ import print_function
import time
import jwt
from urllib.parse import urljoin
import asyncio
import aiohttp
import dateutil.parser
from enum import Enum
from .query_results import YandexQueryResults
from typing import Optional, Callable, Any, Dict
from aiohttp_retry import RandomRetry, RetryClient
from datetime import datetime


class YandexQueryException(Exception):
    """Specific exception for YQ query execution"""

    def __init__(self, issues):
        self.issues = issues

    def __str__(self) -> str:
        return super().__str__()


class YandexQuery():
    """Execute queries in YQ"""

    def __init__(self,
                 base_api_url: str = "https://api.yandex-query.cloud.yandex.net/api/",  # noqa: E501
                 base_iam_url: str = "https://iam.api.cloud.yandex.net",
                 base_vm_metadata_url: str = "http://169.254.169.254/computeMetadata/v1/"):  # noqa: E501
        self.service_account_key = None
        self.base_api_url = base_api_url
        self.base_iam_url = base_iam_url
        self.base_vm_metadata_url = base_vm_metadata_url
        self.auth_type = YandexQuery.AuthType.VM

    # https://cloud.yandex.ru/ru/docs/compute/operations/vm-info/get-info
    async def _resolve_vm_account_key(self) -> str:
        """Resolves IAM token in current VM"""

        url = urljoin(self.base_vm_metadata_url, 'instance/service-accounts/default/token')  # noqa: E501
        headers = {'Metadata-Flavor': 'Google'}

        async with await self.create_async_session(headers=headers) as session:
            async with session.get(url, raise_for_status=True) as response:
                resp = await response.json()
                return resp["access_token"]

    # https://cloud.yandex.ru/ru/docs/compute/operations/vm-info/get-info
    async def resolve_vm_folder_id(self) -> str:
        """Resolves IAM token in current VM"""

        url = urljoin(self.base_vm_metadata_url, 'instance/vendor/?recursive=true')  # noqa: E501
        headers = {'Metadata-Flavor': 'Google'}

        async with await self.create_async_session(headers=headers) as session:
            async with session.get(url, raise_for_status=True) as response:
                resp = await response.json()
                return resp["folderId"]

    # https://cloud.yandex.com/en/docs/iam/operations/iam-token/create-for-sa#get-iam-token
    async def _resolve_service_account_key(self, sa_info: Dict[str,str] ) -> str:
        """Resolves IAM tokey by service account key"""

        async with await self.create_async_session() as session:
            api = urljoin(self.base_iam_url, "iam/v1/tokens")

            now = int(time.time())
            payload = {
                    'aud': api,
                    'iss': sa_info["service_account_id"],
                    'iat': now,
                    'exp': now + 360}

            # Creating the JWT token
            encoded_token = jwt.encode(
                payload,
                sa_info["private_key"],
                algorithm='PS256',
                headers={'kid': sa_info["id"]})

            data = {"jwt": encoded_token}

            async with session.post(api,
                                    json=data,
                                    raise_for_status=True) as response:
                resp = await response.json()
                return resp["iamToken"]

    async def _get_iam_token(self) -> str:
        """Obtains new IAM account"""

        if self.auth_type == YandexQuery.AuthType.VM:
            return await self._resolve_vm_account_key()
        else:
            return await self._resolve_service_account_key(
                self.service_account_key)

    @staticmethod
    def get_request_url_header_params(
            iam_token: Optional[str] = None,
            headers: Optional[dict[str, str]] = None) -> dict[str, str]:

        _headers = {} if headers is None else headers
        if iam_token is not None:
            _headers['Authorization'] = f"Bearer {iam_token}"

        _headers["User-Agent"] = "Jupyter yandex_query_magic"
        return _headers

    def set_service_account_key_auth(self, auth_info: Dict[str, str]):
        """Sets auth mode as service account key file"""

        self.service_account_key = auth_info
        self.auth_type = YandexQuery.AuthType.SA_KEY_FILE

    def set_vm_auth(self):
        """Sets auth mode as VM auth"""

        self.service_account_key = None
        self.auth_type = YandexQuery.AuthType.VM

    class AuthType(Enum):
        """Yandex cloud authorization type"""

        SA_KEY_FILE = 1
        VM = 2

    async def start_execute_query(self,
                                  folder_id: str,
                                  query_text: Optional[str] = None,
                                  name: Optional[str] = None,
                                  description: Optional[str] = None) -> str:
        """Executing query in YQ using current settings
           :return: query id
        """

        type = "ANALYTICS"

        iam_token = await self._get_iam_token()
        async with await self.create_async_session(iam_token) as session:
            data = {"name": name,
                    "type": type,
                    "text": query_text,
                    "description": description}

            url = urljoin(self.base_api_url, f"fq/v1/queries?project={folder_id}") # noqa
            async with session.post(url,
                                    json=data,
                                    raise_for_status=True) as response:
                resp = await response.json()
                self.query_id = resp["id"]

        return self.query_id

    async def create_async_session(self,
                                   iam_token: Optional[str] = None,
                                   headers: Optional[dict[str, str]] = None)\
            -> aiohttp.ClientSession:
        """Creates retriable asyncio session"""

        headers = YandexQuery.get_request_url_header_params(iam_token, headers)
        session = aiohttp.ClientSession(headers=headers,
                                        timeout=aiohttp.ClientTimeout(total=1))

        retriable_exceptions = {
            asyncio.exceptions.TimeoutError,
            asyncio.TimeoutError
            }

        retry_options = RandomRetry(attempts=3,
                                    exceptions=retriable_exceptions)

        client = RetryClient(session, retry_options=retry_options)
        return client

    async def _get_query_status(self, folder_id: str, query_id: str) -> str:
        """Retrieves the query status
           :return: status of the query
        """

        url = urljoin(self.base_api_url,
                      f"fq/v1/queries/{query_id}/status?project={folder_id}")

        iam_token = await self._get_iam_token()
        async with await self.create_async_session(iam_token) as session:
            async with session.get(url, raise_for_status=True) as response:
                resp = await response.json()
                return resp["status"]

    async def wait_results(self,
                           folder_id: str,
                           query_id: str,
                           on_status_update: Callable[[str], None],
                           on_progress_update: Callable[[int, datetime], None]) -> str:  # noqa
        """Wait the query to complete i.e. any status other
        than RUNNING. PENDING
        Reports current status and progress while waiting"""

        progress = 0
        query_info = await self.get_queryinfo(folder_id,
                                              query_id,
                                              await self._get_iam_token())

        started = dateutil.parser.isoparse(query_info["meta"]["started_at"])\
            .replace(tzinfo=None)

        while True:
            status = await self._get_query_status(folder_id, query_id)
            on_status_update(status)
            if status not in ["RUNNING", "PENDING"]:
                progress = 100
                on_progress_update(progress, started)
                return

            await asyncio.sleep(1)
            progress = progress+1
            on_progress_update(progress, started)
            if progress > 100:
                progress = 0

    # https://cloud.yandex.com/en/docs/query/api/methods/get-query
    async def get_queryinfo(self,
                            folder_id: str,
                            query_id: str,
                            iam_token: Optional[str] = None) -> Any:
        """ Retrieves the query execution information
            when query execution is finished already
            :return:
        """
        if iam_token is None:
            iam_token = await self._get_iam_token()

        async with await self.create_async_session(iam_token) as session:
            url = urljoin(self.base_api_url,
                          f"fq/v1/queries/{query_id}?project={folder_id}")

            async with session.get(url, raise_for_status=True) as response:
                resp = await response.json()
                return resp

    # https://cloud.yandex.com/en/docs/query/api/methods/get-query-results
    async def _query_results(self,
                             folder_id: str,
                             query_id: str,
                             result_set_count: int,
                             iam_token: str) -> YandexQueryResults:
        """Retrieves query execution results
        :result_set_count Maximum result set count to retrieve
        :return: YandexQueryResults wrapper over raw results
        """
        results = list()

        async with await self.create_async_session(iam_token) as session:

            # reading results page by page with maximum page size = 1000
            # (as YQ current limit)
            for result_index in range(0, result_set_count):
                limit = 1000
                offset = 0

                columns = None
                rows = []

                while True:
                    url = urljoin(self.base_api_url,
                                  f"fq/v1/queries/{query_id}/"
                                  f"results/{result_index}"
                                  f"?project={folder_id}&"
                                  f"limit={limit}&offset={offset}")

                    async with session.get(url, raise_for_status=True)\
                            as response:

                        qresults = await response.json()
                        if columns is None:
                            columns = qresults["columns"]

                        rows.extend(qresults["rows"])

                        if len(qresults["rows"]) != limit:
                            break
                        else:
                            offset += limit

                results.append({"rows": rows, "columns": columns})

        return YandexQueryResults(results)

    async def get_query_result(self, folder_id: str, query_id: str) -> Any:
        """Retrieves all query results"""

        iam_token = await self._get_iam_token()
        query_info = await self.get_queryinfo(folder_id, query_id, iam_token)

        if query_info["status"] in ["FAILED",
                                    "ABORTED_BY_USER",
                                    "ABORTED_BY_SYSTEM"]:
            issues = query_info.get("issues", None)

            raise YandexQueryException(issues)

        result_set_count = len(query_info["result_sets"])

        return await self._query_results(folder_id,
                                         query_id,
                                         result_set_count,
                                         iam_token)

    # https://cloud.yandex.com/en/docs/query/api/methods/stop-query
    async def stop_query(self, folder_id: str, query_id: str) -> None:
        """Stops the query"""

        iam_token = await self._get_iam_token()

        async with await self.create_async_session(iam_token) as session:
            url = urljoin(self.base_api_url,
                          f"fq/v1/queries/{query_id}/stop?project={folder_id}")
            await session.post(url, raise_for_status=True)



from yandex_query_magic import SqlParser
import pandas as pd
from datetime import datetime
import pytest


def test_sqlrender_noop():
    test_str = "select * from a"
    parser = SqlParser()
    assert parser.reformat("select * from a", {}) == test_str


def test_sqlrender_simple():
    test_str = "select * from {{a}}"
    parser = SqlParser()
    assert parser.reformat(test_str, {"a": "T"}) == "select * from T"


def test_sqlrender_df():
    test_str = "select * from {{df}}"
    parser = SqlParser()
    dataframe = pd.DataFrame({'_float': [1.0],
                              '_int': [1],
                              '_datetime': [pd.Timestamp('20180310')],
                              '_string': ['foo']})

    result = parser.reformat(test_str, {"df": dataframe})
    assert result == 'select * from AS_TABLE(AsList(AsStruct(1.0 as `_float`,1l as `_int`,DateTime::MakeTimestamp(DateTime::Parse("%Y-%m-%d %H:%M:%S")("2018-03-10 00:00:00.000000")) as `_datetime`,"foo" as `_string`))) as `df`'  # noqa


def test_sqlrender_df_ms():
    test_str = "select * from {{df}}"
    parser = SqlParser()
    dataframe = pd.DataFrame({'_float': [1.0],
                              '_int': [1],
                              '_datetime': [pd.Timestamp(year=2018,
                                                         month=3,
                                                         day=10,
                                                         hour=1,
                                                         minute=12,
                                                         second=6,
                                                         microsecond=12)],
                              '_string': ['foo']})

    result = parser.reformat(test_str, {"df": dataframe})
    assert result == 'select * from AS_TABLE(AsList(AsStruct(1.0 as `_float`,1l as `_int`,DateTime::MakeTimestamp(DateTime::Parse("%Y-%m-%d %H:%M:%S")("2018-03-10 01:12:06.000012")) as `_datetime`,"foo" as `_string`))) as `df`'  # noqa


def test_sqlrender_df_ns():
    with pytest.raises(Exception, match="No support for nanoseconds of Pandas datetime64"):
        test_str = "select * from {{df}}"
        parser = SqlParser()
        dataframe = pd.DataFrame({'_float': [1.0],
                                  '_int': [1],
                                  '_datetime': [pd.Timestamp(year=2018,
                                                             month=3,
                                                             day=10,
                                                             hour=1,
                                                             minute=12,
                                                             second=6,
                                                             microsecond=12,
                                                             nanosecond=1)],
                                  '_string': ['foo']})
        parser.reformat(test_str,{"df":dataframe})


def test_sqlrender_dict():
    test_str = "select DictHasItems({{a}})"
    a = {"a": 1, "b": 2, "c": 3}
    parser = SqlParser()
    formatted = parser.reformat(test_str, {"a": a})
    print(formatted)
    assert formatted == 'select DictHasItems(ToDict(AsList(asTuple("a", 1l),asTuple("b", 2l),asTuple("c", 3l))))'  # noqa


def test_sqlrender_dict_distinct_value():
    with pytest.raises(Exception, match="All value types must be of one type. Found several {<class 'datetime.datetime'>, <class 'str'>, <class 'int'>, <class 'float'>}"):  # noqa
        test_str = "select * from {{a}}"
        a = {"a": 1, "b": 2.0, "c": "test", "d": datetime(2022, 2, 2, 21, 12, 12)}
        parser = SqlParser()
        formatted = parser.reformat(test_str, {"a": a})
        print(formatted)
        assert formatted == 'select * from AS_TABLE(AsList(AsStruct(1l as `a`,2.0 as `b`,"test" as `c`,DateTime::MakeTimestamp(DateTime::Parse("%Y-%m-%d %H:%M:%S")("2022-02-02 21:12:12.000000")) as `d`))) as `a`'  # noqa


def test_sqlrender_dict_distinct_keys():
    with pytest.raises(Exception, match="All key types must be of one type. Found several {<class 'str'>, <class 'datetime.datetime'>}"):  # noqa
        test_str = "select * from {{a}}"
        a = {"a": 1, "b": 1, "c": 1, datetime.now(): 1}
        parser = SqlParser()
        formatted = parser.reformat(test_str, {"a": a})
        print(formatted)
        assert formatted == 'select * from AS_TABLE(AsList(AsStruct(1l as `a`,2.0 as `b`,"test" as `c`,DateTime::MakeTimestamp(DateTime::Parse("%Y-%m-%d %H:%M:%S")("2022-02-02 21:12:12.000000")) as `d`))) as `a`'  # noqa


def test_sqlrender_list():
    test_str = "select 1 in {{a}}"
    a = [1, 2, 3]
    parser = SqlParser()
    result = parser.reformat(test_str, {"a": a})
    assert result == 'select 1 in AsList(1l,2l,3l)'  # noqa


def test_sqlrender_dict_special_symbols():
    test_str = "select * from {{a}}"
    a = {"a": "abc", "b": "a\"a", "c": "a\'c"}
    parser = SqlParser()
    result = parser.reformat(test_str, {"a": a})
    print(result)
    assert result == """select * from ToDict(AsList(asTuple("a", "abc"),asTuple("b", "a\\"a"),asTuple("c", "a'c")))"""  # noqa



from .main import YandexQuery, YandexQueryException
import argparse
from IPython.core.magic_arguments import (argument,
                                          magic_arguments,
                                          parse_argstring)

from IPython.core.magic import (Magics, magics_class,
                                line_cell_magic, line_magic,
                                no_var_expand)
from IPython.display import display
import re
from jupyter_ui_poll import ui_events
import asyncio
import ipywidgets as widgets
from datetime import datetime
from .jinja_template import JinjaTemplate
from typing import Optional, Dict
from .ipythondisplay import IpythonDisplay
import nest_asyncio
import json
from .sqltext_parser import SqlParser

nest_asyncio.apply()


@magics_class
class YQMagics(Magics):
    """Main class for Jupyter magics interop"""

    DefaultFolderId = None  # default folder to be used between queries
    Sa_info = None  # authentication token to be used between queries

    def __init__(self, shell):
        Magics.__init__(self, shell=shell)
        self.ipython_display = IpythonDisplay()

    # Executes query in YQ
    async def yq_execute_query(self,
                               folder_id: Optional[str],
                               query_text: str,
                               name: Optional[str] = None,
                               description: Optional[str] = None,
                               as_dataframe: bool = True) -> None:

        yq = YandexQuery()
        if YQMagics.Sa_info is not None:
            yq.set_service_account_key_auth(YQMagics.Sa_info)
        else:
            yq.set_vm_auth()

        variable = None

        # parses query <variable> << select ...
        # extracts <variable> to be assigned
        # after query execution to query results
        match = re.search(r'(\b[a-zA-Z_][a-zA-Z0-9_]*\b)\s*<<(.*)',
                          query_text,
                          re.MULTILINE | re.DOTALL)

        if match is not None\
                and len(match.groups()) == 2\
                and str.isidentifier(match[1]):
            variable = match[1]
            query_text = match[2]

        if folder_id is None:
            folder_id = YQMagics.DefaultFolderId

        if folder_id is None:
            try:
                folder_id = await YandexQuery().resolve_vm_folder_id()
            except:
                pass

        if folder_id is None:
            print("Folder id is not specified. "
                  "Specify it with %yq_settings "
                  "--folder_id <folder_id> extension")
            return

        # We use rich interaction with UI and async query execution
        # so we need to run asyncio loops when we need it and interrupt it
        # in required places
        loop = asyncio.get_event_loop()

        query_id = None
        label_query_id = widgets.HTML()

        now_str = datetime.now().strftime('%d.%m.%Y %H:%M:%S')
        # shows start status of query
        start_status = widgets.Label(f"Query started at {now_str}.")

        # shows final information about query
        stop_status = widgets.Label("")

        # shows final information about query
        is_truncated_label = widgets.Label("")
        is_truncated_label.style.text_color = 'violet'

        # aborts execution of the query
        abort_query_button = widgets.Button(description="Abort")

        # shows the progress of query execution
        # It is not the best control to show progress of unknown length
        # But ipywidgets library has not another control
        progress = widgets.IntProgress(
            value=0,
            min=0,
            max=100,
            bar_style='info',
            style={'bar_color': 'green'},
            orientation='horizontal'
        )

        # Shows progress of query execution
        processed_gb = widgets.Label("")  # 0 GB

        # Hidden field for query execution issues
        issues = widgets.Textarea()
        issues.layout = widgets.Layout(width='600px', height="300px")
        issues.layout.display = 'none'
        issues.style = dict(text_color='red')

        # UI control layout
        all_widgets = widgets.VBox(
            [widgets.HBox([label_query_id, is_truncated_label]),
             widgets.HBox(
                 [start_status, stop_status]),
             widgets.HBox(
                 [progress, processed_gb, abort_query_button]),
             issues])

        # Update status callback from start_execute_query method
        def update_status(new_status: str) -> None:
            progress.description = new_status

        # Update progress callback from start_execute_query method
        def update_progress(new_value: int, query_started_at: datetime) -> None: # noqa
            progress.value = new_value
            query_execution_time = datetime.now().replace(tzinfo=None) - query_started_at # noqa
            stop_status.value = f"Executing for {query_execution_time} "\
                                "seconds so far"

            # Singe we need to react on Abort button,
            # we need to process UI event loop from time to time
            with ui_events() as ui_poll:
                ui_poll(1)

        async def abort_query_async(query_id_: str) -> None:
            try:
                await yq.stop_query(folder_id, query_id_)
            except Exception as stop_ex:
                stop_status.value = str(stop_ex)

        # Callback to stop the query
        def abort_query(_):
            loop.create_task(abort_query_async(query_id))

        abort_query_button.on_click(abort_query)

        # Show all controls
        display(all_widgets)  # noqa

        started_at = datetime.now()
        query_id = await yq.start_execute_query(
            folder_id,
            query_text,
            name,
            description)

        label_query_id.value = f"Query id is <a style='text-decoration: underline;'"\
                               f" href='https://yq.cloud.yandex.ru/folders/{folder_id}/ide/queries/{query_id}'"\
                               f" target='_blank'>{query_id}</a>." # noqa

        result = None
        try:
            # Start query execution
            await yq.wait_results(
                    folder_id,
                    query_id,
                    update_status,
                    update_progress)

            try:
                query_info = await yq.get_queryinfo(
                                folder_id,
                                query_id)

                query_status = query_info["status"]
                progress.description = query_status

                for resultSet in query_info["result_sets"]:
                    is_truncated_rs = resultSet.get("truncated", False)
                    if is_truncated_rs:
                        is_truncated_label.value = "Results were truncated";

                progress.description = "Fetching results"

                # Retrieving query results
                try:
                    if query_status == "COMPLETED":
                        result = await yq.get_query_result(folder_id, query_id)
                        progress.description = "DONE"
                        progress.bar_style = "success"

                    else:
                        progress.value = 0
                        progress.description = query_status
                        progress.bar_style = "danger"

                        if "issues" in query_info:
                            issues.value = json.dumps(query_info["issues"])
                            issues.layout.display = 'block'

                # If YQ query execution happened, show details
                except YandexQueryException as ex:
                    issues.value = ex.issues.__repr__()
                    issues.layout.display = 'block'

            except Exception as ex:
                self.ipython_display.error(ex.__repr__())
                # issues.value = ex.__repr__()
                # issues.layout.display = 'block'

        finally:
            # Hide abort query button after query execution completed
            abort_query_button.layout.display = 'none'

        total_time = str(datetime.now()-started_at)
        finish_time_str = datetime.now().strftime('%d.%m.%Y %H:%M:%S')
        stop_status.value = f"Finished at {finish_time_str}."\
                            f" Total time is {total_time}"

        if result is not None:
            if as_dataframe:
                result = result.to_dataframe()
            else:
                result = result.raw_results

        # Write results to external variable
        if variable is not None:
            self.shell.user_ns[variable] = result
            return result
        else:
            return result

    @no_var_expand
    @line_magic
    @magic_arguments()
    @argument("--sa-file-auth", help="Authenticate using path to Yandex Cloud static credentials file", type=str)  # noqa
    @argument("--vm-auth", help="Authenticate use VM credentials", action="store_true")  # noqa
    @argument("--folder-id", help="Yandex cloud folder id to run queries", type=str)  # noqa
    def yq_settings(self, line):
        args = parse_argstring(self.yq_settings, line)

        if args.vm_auth:
            YQMagics.SA_info = None
        # read file with SA credentials
        elif args.sa_file_auth is not None:
            sa_file = args.sa_file_auth.strip()
            with open(sa_file, "r") as sa_file:
                sa_info = sa_file.read()
            YQMagics.Sa_info = json.loads(sa_info)

        if args.folder_id is not None:
            YQMagics.DefaultFolderId = args.folder_id

    @no_var_expand
    @magic_arguments()
    @line_cell_magic("yq")
    @argument("--folder-id", help="Yandex cloud folder id to run queries", type=str)  # noqa
    @argument("--name", help="Query name", type=str)
    @argument("--description", help="Query description", type=str)
    @argument("-j", "--jinja2", help="Apply Jinja2 Template", action="store_true")  # noqa
    @argument("--no-var-expansion", help="Disable {{var}} evaluation", action="store_true")  # noqa
    @argument("--raw-results", help="Return result as raw YQ response", action='store_true', default=False)  # noqa
    @argument("rest", nargs=argparse.REMAINDER)
    def execute(self, line: Optional[str] = None,
                cell: Optional[str] = None) -> None:

        cell = "" if cell is None else cell
        line = "" if line is None else line
        user_ns: Dict[str, object] = self.shell.user_ns

        args = parse_argstring(self.execute, line)

        rest = " ".join(args.rest)
        query = f"{rest}\n{cell}".strip()

        if args.jinja2:
            query = JinjaTemplate.apply_template(query, user_ns)
        elif args.no_var_expansion:
            pass
        else:
            parser = SqlParser()
            query = parser.reformat(query, user_ns)

        loop = asyncio.get_event_loop()

        query_result = loop.run_until_complete(
            self.yq_execute_query(args.folder_id,
                                  query, args.name,
                                  args.description,
                                  not args.raw_results))
        return query_result


def load_ipython_extension(ip):
    """Load the extension in IPython."""
    ip.register_magics(YQMagics)


def unload_ipython_extension(ip):
    """Unload the extension in IPython."""
    if 'YandexQuery' in ip.magics_manager.registry:
        del ip.magics_manager.registry['YQMagics']
    if 'YQMagics' in ip.config:
        del ip.config['YQMagics']



import pytest
from yandex_query_magic import YandexQuery, YandexQueryException
from pytest_httpserver import HTTPServer, httpserver
from werkzeug.wrappers import Response
import datetime
import time
import jwt
import json


# Test sa key to be used in tests
TEST_SA_KEY = {
  "id": "ajer7vgr0lsf92k6ftbf",
  "service_account_id": "aje5cgv2rif15p739hve",
  "created_at": "2024-02-01T08:32:16.871406224Z",
  "key_algorithm": "RSA_2048",
  "public_key": "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAvN76kfg8U4itnysDyqGm\nt6xgxxFFnWSGOMCFtGSnYjchs11h8asNHLLSdfqRWSocoHSKJw5Z598sAoUlLbYC\nOxlz/lkdXV5lU+b9jUfTtl/9Q5ior9xurhzGYQCmaYY0/9UBRMFDTSzrW9rjaM6H\n3KaSZs3WRvyB8nGqk3f6DruN6z1wR0LtrvoTMVvDs8mLzs/K4VjqTziLWGfCuh0d\n2KGmGjyWSwN+XRadjqd+DjEGA3CaVAgLzqzY25zpOYzZYYsKRAB58/zl9KEAhN3b\nFYeBSJ359RbL7SJ8KvC8jmWvZZUmT8eP77gvd6l1Xp3Kr/dB9Rxx2SfZTVKtNUDV\nlQIDAQAB\n-----END PUBLIC KEY-----\n",  # noqa
  "private_key": "PLEASE DO NOT REMOVE THIS LINE! Yandex.Cloud SA Key ID <ajer7vgr0lsf92k6ftbf>\n-----BEGIN PRIVATE KEY-----\nMIIEvgIBADANBgkqhkiG9w0BAQEFAASCBKgwggSkAgEAAoIBAQC83vqR+DxTiK2f\nKwPKoaa3rGDHEUWdZIY4wIW0ZKdiNyGzXWHxqw0cstJ1+pFZKhygdIonDlnn3ywC\nhSUttgI7GXP+WR1dXmVT5v2NR9O2X/1DmKiv3G6uHMZhAKZphjT/1QFEwUNNLOtb\n2uNozofcppJmzdZG/IHycaqTd/oOu43rPXBHQu2u+hMxW8OzyYvOz8rhWOpPOItY\nZ8K6HR3YoaYaPJZLA35dFp2Op34OMQYDcJpUCAvOrNjbnOk5jNlhiwpEAHnz/OX0\noQCE3dsVh4FInfn1FsvtInwq8LyOZa9llSZPx4/vuC93qXVencqv90H1HHHZJ9lN\nUq01QNWVAgMBAAECggEAUKKyP9fHibJ0zdvDhqN1Vj2WI+dP3V6pn1kyvE2s2NXI\n4Zcg1di9hF8kU5Jis7qy9h5LTVlnMQOq+nh14wot8aVwTEsnqlE+2Y9o+QSNcvOQ\nYWevvUVTS6qlV4y7f5n4zrDWFdCdNznSUiklpf1nK+FB5/pBXZU4tZWpycQTUm4r\nnt+Zgu9hJHe0vDDw3VQEyNswd38zFELj0NdPZWgpsEdOxBfL5eaSkNF/oXPnO8Pu\nkLNus0klOcaKG9IzoTy2wf4587n5FcyeViRi3pRl7JM+2m9LtudMiS9YrNsx0vAI\nvkahGpFWFEnjg0Bzrz09mO+yqgX+k+iw88EJcedQkQKBgQD3S7BwDEugOCprF/I3\n3clLaGgvmWxDegKu3FsWFWOROKpsdaM09NBZVTqB20nBwKURy/+JugDyLMKzCDwH\nRiC5ROsyxoISAQAmdxspGHYQ4bQEoUQURR+UPBvPOjXawx7Qczv+6+0foy2MAmYq\n+KC8gC88ov7g2Q4llIWUIc74NwKBgQDDhNd5Mt6r9d7Cq4fy+9W0sveRvbLN0G9I\nOvNPYuPKiz4M2l/Gli+zQSJodJGpRnRTxFxRyy3/qHe+rBHfnR9FRSpdMyowB7j3\n00hZK9qg3/hrSoUiyQHovwHmncO2zxpC6PwqRTLjmmKHhg+8AATqjqq/EwPn8FVi\ngf8USpcikwKBgQCX1Euub44a/4DjoZ8gN7Y36xFUcCDtSMLO8xGlfFpWNfFEh/E9\nOTWWM2KpiaY+I/X0+EebGq0sAtlDLEIWwTKkLTEuSnxfa2fZNfViBNewQ7LPyOdQ\nfqQF0eXNFFMuTe/kUWu7dsRuUTRMqshph57APP9Dflt4Vyt9XTOqIBu/fwKBgQCS\nhszg9MGyB9qA9AI3lIpNGM62t374BZxQenV76jWixoWjJkkez9FBuq+prqq3PKjT\nWlaBqg54Ce5rxBLFDcCwriwYms6kHjV97SbMRTRc2l0XM0rhjdjTb8ph5ZwWNdGN\nkYPhvehscHgk7tB96VnF21OVTQ/hU9j5sUjUES6A5wKBgHvzNT6bYhURpymppktQ\nii/wSCaW+JKB+re3HirTkQkg7XyP1tdJmjC8u0XznWW32oIUu3Yc5u3NpM2UyFhu\nvsbJmXl/Ku3Jh24HTRV+ZC3TDNcMZXHCBaiZkfTYXSpWwOVTETu/AWe9Qa1lBGTx\njIX5w9L1zoLytMGlZ3lrcCKN\n-----END PRIVATE KEY-----\n"  # noqa
}


@pytest.fixture(scope="function")
def yandex_query(yq_httpserver: HTTPServer,
                 iam_httpserver: HTTPServer,
                 vm_httpserver: HTTPServer) -> YandexQuery:

    base_api_url = yq_httpserver.url_for("/yq")
    base_iam_url = iam_httpserver.url_for("/iam")
    base_vm_url = vm_httpserver.url_for("/vm")

    yq = YandexQuery(base_api_url=base_api_url,
                     base_iam_url=base_iam_url,
                     base_vm_metadata_url=base_vm_url)

    yq.set_service_account_key_auth(TEST_SA_KEY)
    return yq


@pytest.fixture(scope="function")
def yq_httpserver() -> HTTPServer:
    """Main YQ HTTP-server"""

    server = HTTPServer()
    server.start()
    yield server
    server.stop()


@pytest.fixture(scope="function")
def vm_httpserver() -> HTTPServer:
    """Main YQ HTTP-server"""

    server = HTTPServer()
    server.start()
    yield server
    server.stop()


@pytest.fixture(scope="function")
def iam_httpserver() -> HTTPServer:
    """IAM HTTP-server"""

    server = HTTPServer()
    server.start()

    yield server
    server.stop()


@pytest.mark.asyncio
async def test_auth(iam_httpserver: HTTPServer, yandex_query: YandexQuery):
    """Tests Yandex Cloud IAM integration"""

    assert yandex_query is not None
    api = iam_httpserver.url_for("/iam/v1/tokens")

    # Custom handler to check correctness of IAM data
    def handler(r):
        decode_result = jwt.decode(
            json.loads(r.data)["jwt"],
            TEST_SA_KEY["public_key"],
            algorithms=["PS256"], audience=api)

        assert decode_result["iss"] == TEST_SA_KEY["service_account_id"]

        return Response(json.dumps({"iamToken": "iamToken"}),
                        status=200,
                        content_type="application/json")

    iam_httpserver.expect_request("/iam/v1/tokens",
                                  method="POST").\
                                  respond_with_handler(handler)  # noqa

    await yandex_query._get_iam_token()


@pytest.mark.asyncio
async def test_auth_timeout(iam_httpserver: HTTPServer,
                            yandex_query: YandexQuery):
    """Tests Yandex Cloud IAM timeouts"""

    assert yandex_query is not None

    def sleeping(_):
        time.sleep(2)
        return Response(json.dumps({"iamToken": "iamToken"}),
                        status=200, content_type="application/json")

    iam_httpserver.expect_oneshot_request("/iam/v1/tokens",
                                          method="POST").respond_with_handler(
                                              sleeping)

    iam_httpserver.expect_request("/iam/v1/tokens",
                                  method="POST").respond_with_json(
                                      {"iamToken": "test_iam_token"})

    await yandex_query._get_iam_token()


@pytest.mark.asyncio
async def test_user_agent(iam_httpserver: HTTPServer,
                          yq_httpserver: HTTPServer,
                          yandex_query: YandexQuery):
    """Tests happy path to create query"""

    assert yandex_query is not None

    folder_id = "folder_id"
    query_id = "query_id"
    name = "name"
    description = "description"
    query_text = "select 1"

    # Set up IAM to handle all auth queries
    iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                  headers={"User-Agent": "Jupyter yandex_query_magic"},  # noqa
                                  handler_type=httpserver.HandlerType.PERMANENT).\
        respond_with_json({"iamToken": "test_iam_token"})  # noqa

    # Checking correctness of create-query command
    yq_httpserver.expect_request("/fq/v1/queries",
                                 query_string=f"project={folder_id}",
                                 method="POST",
                                 headers={"User-Agent": "Jupyter yandex_query_magic"},  # noqa
                                 json={"name": name,
                                       "description": description,
                                       "text": query_text,
                                       "type": "ANALYTICS"}).\
        respond_with_json({"id": query_id})

    # Issue create-query command
    expected_query_id = await yandex_query.start_execute_query(folder_id,
                                                               query_text,
                                                               name,
                                                               description)

    # Checking if expected query id is the same as created
    assert expected_query_id == query_id


@pytest.mark.asyncio
async def test_query(iam_httpserver: HTTPServer,
                     yq_httpserver: HTTPServer,
                     yandex_query: YandexQuery):
    """Tests happy path to create query"""

    assert yandex_query is not None

    folder_id = "folder_id"
    query_id = "query_id"
    name = "name"
    description = "description"
    query_text = "select 1"

    # Set up IAM to handle all auth queries
    iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                  handler_type=httpserver.HandlerType.PERMANENT).\
        respond_with_json({"iamToken": "test_iam_token"})  # noqa

    # Checking correctness of create-query command
    yq_httpserver.expect_request("/fq/v1/queries",
                                 query_string=f"project={folder_id}",
                                 method="POST",
                                 json={"name": name,
                                       "description": description,
                                       "text": query_text,
                                       "type": "ANALYTICS"}).\
        respond_with_json({"id": query_id})

    # Issue create-query command
    expected_query_id = await yandex_query.start_execute_query(folder_id,
                                                               query_text,
                                                               name,
                                                               description)

    # Checking if expected query id is the same as created
    assert expected_query_id == query_id


@pytest.mark.asyncio
async def test_get_query_status(iam_httpserver: HTTPServer,
                                yq_httpserver: HTTPServer,
                                yandex_query: YandexQuery):
    """Tests get-query-status happy path command"""

    assert yandex_query is not None

    folder_id = "folder_id"
    query_id = "query_id"
    status = "RUNNING"

    # Set up IAM to handle all auth queries
    iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                  handler_type=httpserver.HandlerType.PERMANENT).\
        respond_with_json({"iamToken": "test_iam_token"})  # noqa

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/status",
                                 query_string=f"project={folder_id}",
                                 method="GET").\
        respond_with_json({"status": status})

    # Check if returned query status is the same as expected
    expected_status = await yandex_query._get_query_status(folder_id, query_id)
    assert expected_status == status


@pytest.mark.asyncio
async def test_get_query_info(iam_httpserver: HTTPServer,
                              yq_httpserver: HTTPServer,
                              yandex_query: YandexQuery):
    """Tests happy path for get-query-info command"""

    assert yandex_query is not None

    folder_id = "folder_id"
    query_id = "query_id"
    status = "RUNNING"

    # Set up IAM to handle all auth queries
    iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                  handler_type=httpserver.HandlerType.PERMANENT).\
        respond_with_json({"iamToken": "test_iam_token"})  # noqa

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}",
                                 query_string=f"project={folder_id}",
                                 method="GET").\
        respond_with_json({"status": status})

    # Checks if response is the same as expected
    query_info = await yandex_query.get_queryinfo(folder_id, query_id)
    assert query_info["status"] == status


@pytest.mark.asyncio
async def test_get_query_status_timedout(iam_httpserver: HTTPServer,
                                         yq_httpserver: HTTPServer,
                                         yandex_query: YandexQuery):
    """Tests timeout for get-query-status command"""

    assert yandex_query is not None

    folder_id = "folder_id"
    query_id = "query_id"
    status = "FAILED"

    def sleeping(_):
        time.sleep(2)
        return Response(json.dumps({"status": status}),
                        status=200,
                        content_type="application/json")

    # Set up IAM to handle all auth queries
    iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                  handler_type=httpserver.HandlerType.PERMANENT).\
        respond_with_json({"iamToken": "test_iam_token"})  # noqa

    # First query responds with timeout 2 seconds
    yq_httpserver.expect_oneshot_request(f"/fq/v1/queries/{query_id}/status",
                                         query_string=f"project={folder_id}",
                                         method="GET").\
        respond_with_handler(sleeping)

    # Second query responds with HTTP 500 error code
    yq_httpserver.expect_oneshot_request(f"/fq/v1/queries/{query_id}/status",
                                         query_string=f"project={folder_id}",
                                         method="GET").\
        respond_with_response(Response(status=500))

    # Finally query responds HTTP 200 OK
    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/status",
                                 query_string=f"project={folder_id}",
                                 method="GET").\
        respond_with_json({"status": status})

    # Query must retry all that errors
    expected_status = await yandex_query._get_query_status(folder_id, query_id)
    assert expected_status == status


@pytest.mark.asyncio
async def test_wait_results_complete(iam_httpserver: HTTPServer,
                                     yq_httpserver: HTTPServer,
                                     yandex_query: YandexQuery):
    """Tests happy path for long wait_results process"""

    assert yandex_query is not None

    folder_id = "folder_id"
    query_id = "query_id"
    started_at = datetime.datetime.now().isoformat()

    def status_response():
        return {"meta": {"started_at": started_at}}

    # Set up IAM to handle all auth queries
    iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                  handler_type=httpserver.HandlerType.PERMANENT).\
        respond_with_json({"iamToken": "test_iam_token"})  # noqa

    # Get ready to several queries in specific order
    yq_httpserver.expect_ordered_request(f"/fq/v1/queries/{query_id}",
                                         query_string=f"project={folder_id}",
                                         method="GET").\
        respond_with_json(status_response())

    yq_httpserver.expect_ordered_request(f"/fq/v1/queries/{query_id}/status",
                                         query_string=f"project={folder_id}",
                                         method="GET").\
        respond_with_json({"status": "RUNNING"})

    yq_httpserver.expect_ordered_request(f"/fq/v1/queries/{query_id}/status",
                                         query_string=f"project={folder_id}",
                                         method="GET").\
        respond_with_json({"status": "SUCCESS"})

    status = None

    def status_callback(_status):
        nonlocal status
        status = _status

    progress = 0

    def progress_callback(_progress, __):
        nonlocal progress
        progress = _progress

    await yandex_query.wait_results(folder_id,
                                    query_id,
                                    status_callback,
                                    progress_callback)

    # Checking final query status and final progress
    assert status == "SUCCESS"
    assert progress == 100


@pytest.mark.asyncio
async def test_get_query_result_bad_status(iam_httpserver: HTTPServer,
                                           yq_httpserver: HTTPServer,
                                           yandex_query: YandexQuery):
    """Checks that get-query-result throws exception if query execution failed"""

    assert yandex_query is not None

    with pytest.raises(YandexQueryException):
        folder_id = "folder_id"
        query_id = "query_id"
        status = "FAILED"

        # Set up IAM to handle all auth queries
        iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                      handler_type=httpserver.HandlerType.PERMANENT).\
            respond_with_json({"iamToken": "test_iam_token"})  # noqa

        yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}",
                                     query_string=f"project={folder_id}",
                                     method="GET").\
            respond_with_json({"status": status})

        await yandex_query.get_query_result(folder_id, query_id)


@pytest.mark.asyncio
async def test_get_query_results(iam_httpserver: HTTPServer,
                                 yq_httpserver: HTTPServer,
                                 yandex_query: YandexQuery):
    """Checks corectness of simple get-query-result command"""
    assert yandex_query is not None

    def result_response():
        return {
            "columns": [{"name": "a", "type": "int"}],
            "rows": [[0]]
                }

    folder_id = "folder_id"
    query_id = "query_id"

    # Set up IAM to handle all auth queries
    iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                  handler_type=httpserver.HandlerType.PERMANENT).\
        respond_with_json({"iamToken": "test_iam_token"})  # noqa

    # Respond
    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/0",
                                 query_string=f"project={folder_id}&limit=1000&offset=0",  # noqa
                                 method="GET").\
        respond_with_json(result_response())

    iam_token = await yandex_query._get_iam_token()
    result = await yandex_query._query_results(folder_id,
                                               query_id, 1,
                                               iam_token)
    assert len(result.results[0]["rows"]) == 1
    assert result.results[0]["rows"][0][0] == 0


@pytest.mark.asyncio
async def test_get_query_big_results(iam_httpserver: HTTPServer,
                                     yq_httpserver: HTTPServer,
                                     yandex_query: YandexQuery):
    """Tests multipage response and limit-offset listing of results"""

    assert yandex_query is not None

    def result_response(offset):
        response = {
            "columns": [{"name": "a", "type": "int"}],
            "rows": [[value] for value in range(offset, 2500)[0:1000]]
                }
        return response

    folder_id = "folder_id"
    query_id = "query_id"

    # Set up IAM to handle all auth queries
    iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                  handler_type=httpserver.HandlerType.PERMANENT).\
        respond_with_json({"iamToken": "test_iam_token"})  # noqa

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/0",
                                 query_string=f"project={folder_id}&limit=1000&offset=0",  # noqa
                                 method="GET").\
        respond_with_json(result_response(0))

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/0",
                                 query_string=f"project={folder_id}&limit=1000&offset=1000",  # noqa
                                 method="GET").\
        respond_with_json(result_response(1000))

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/0",
                                 query_string=f"project={folder_id}&limit=1000&offset=2000",  # noqa
                                 method="GET").\
        respond_with_json(result_response(2000))

    iam_token = await yandex_query._get_iam_token()
    result = await yandex_query._query_results(folder_id, query_id,
                                               1, iam_token)

    # Checks if results len is expected
    assert len(result.results[0]["rows"]) == 2500

    # Checks if results are as expected
    assert result.results[0]["rows"] == [[value] for value in range(0, 2500)]


@pytest.mark.asyncio
async def test_get_query_result_multiple_resultsets(iam_httpserver: HTTPServer,
                                                    yq_httpserver: HTTPServer,
                                                    yandex_query: YandexQuery):
    """Tests multipage response, limit-offset and multi resultset response"""
    assert yandex_query is not None

    started_at = datetime.datetime.now().isoformat()

    def status_response():
        return {
                "meta": {
                    "started_at": started_at},
                "result_sets": [
                    {
                        "rows": 2500, "truncated": False
                    },
                    {
                        "rows": 3999, "truncated": False
                    }
                ],
                "status": "COMPLETED"}

    folder_id = "folder_id"
    query_id = "query_id"

    # Set up IAM to handle all auth queries
    iam_httpserver.expect_request("/iam/v1/tokens", method="POST",
                                  handler_type=httpserver.HandlerType.PERMANENT).\
        respond_with_json({"iamToken": "test_iam_token"})  # noqa

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}",
                                 query_string=f"project={folder_id}",
                                 method="GET").\
        respond_with_json(status_response())

    def result_response0(offset):
        response = {
            "columns": [{"name": "a", "type": "int"}],
            "rows": [[value] for value in range(offset, 2500)[0:1000]]
                }
        return response

    def result_response1(offset):
        response = {
            "columns": [{"name": "b", "type": "int"}],
            "rows": [[value] for value in range(offset + 999, 3999+999)[0:1000] ]  # noqa
                }
        return response

    folder_id = "folder_id"
    query_id = "query_id"

    # Setting up query sequence
    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/0",
                                 query_string=f"project={folder_id}&limit=1000&offset=0",  # noqa
                                 method="GET").\
        respond_with_json(result_response0(0))

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/0",
                                 query_string=f"project={folder_id}&limit=1000&offset=1000",  # noqa
                                 method="GET").\
        respond_with_json(result_response0(1000))

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/0",
                                 query_string=f"project={folder_id}&limit=1000&offset=2000",  # noqa
                                 method="GET").\
        respond_with_json(result_response0(2000))

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/1",
                                 query_string=f"project={folder_id}&limit=1000&offset=0",  # noqa
                                 method="GET").\
        respond_with_json(result_response1(0))

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/1",
                                 query_string=f"project={folder_id}&limit=1000&offset=1000",  # noqa
                                 method="GET").\
        respond_with_json(result_response1(1000))

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/1",
                                 query_string=f"project={folder_id}&limit=1000&offset=2000",  # noqa
                                 method="GET").\
        respond_with_json(result_response1(2000))

    yq_httpserver.expect_request(f"/fq/v1/queries/{query_id}/results/1",
                                 query_string=f"project={folder_id}&limit=1000&offset=3000",  # noqa
                                 method="GET").\
        respond_with_json(result_response1(3000))

    result = await yandex_query.get_query_result(folder_id, query_id)

    # Checks the results of first result set
    assert len(result.results[0]["rows"]) == 2500
    assert result.results[0]["rows"] == [[value] for value in range(0, 2500)]

    # Checks the results of second result set
    assert len(result.results[1]["rows"]) == 3999
    assert result.results[1]["rows"] == [[value] for value in range(999, 3999+999)]  # noqa


@pytest.mark.asyncio
async def test_vm_auth(iam_httpserver: HTTPServer,
                       yandex_query: YandexQuery,
                       vm_httpserver: HTTPServer):
    """Tests Yandex Cloud VM authentication"""

    assert yandex_query is not None
    vm_httpserver.expect_request("/instance/service-accounts/default/token",
                                 method="GET",
                                 headers={"Metadata-Flavor": "Google"}).\
        respond_with_json({"access_token": "test_iam_token"})

    yandex_query.set_vm_auth()
    token = await yandex_query._get_iam_token()
    assert token == "test_iam_token"


@pytest.mark.asyncio
async def test_auth_switch(iam_httpserver: HTTPServer,
                       yandex_query: YandexQuery,
                       vm_httpserver: HTTPServer):
    """Tests switching between Yandex Cloud VM and IAM authentication"""

    yandex_query.set_vm_auth()
    await test_vm_auth(iam_httpserver, yandex_query, vm_httpserver)

    yandex_query.set_service_account_key_auth(TEST_SA_KEY)
    await test_auth(iam_httpserver, yandex_query)

    yandex_query.set_vm_auth()
    await test_vm_auth(iam_httpserver, yandex_query, vm_httpserver)



from typing import Dict, Any, Optional
from .sqltext_parser import SqlParser


class JinjaTemplate:

    @staticmethod
    def apply_template(sql: str, user_ns: Dict[str, object]) -> str:
        try:
            from jinja2 import Template, Environment
        except Exception as e:
            raise ValueError(
                "Jinja2 must be installed to use --jinja2: %pip3 install Jinja2"
            ) from e

        def to_yq(value: Any, name: Optional[str] = None) -> Any:
            return SqlParser.render_type(value, name)

        env = Environment()
        env.filters["to_yq"] = to_yq

        t = env.from_string(sql)
        return t.render(user_ns)



from IPython import get_ipython
import sys


class IpythonDisplay(object):
    def __init__(self):
        self._ipython_shell = get_ipython()

    def stderr_flush(self):
        sys.stderr.flush()

    def error(self, error):
        sys.stderr.write("{}\n".format(error))
        self.stderr_flush()



import re
from datetime import datetime
import pandas as pd
from typing import Any


class SqlParser:
    mustache_re = re.compile(r"{{\s*([a-zA-Z_][a-zA-Z0-9_]*)(.*)\s*}}")

    def __init__(self):
        pass

    def reformat(self, sql, ns):
        new_sql = ""
        prev_position = 0

        while True:
            match = SqlParser.mustache_re.search(sql, prev_position)
            if match is None:
                new_sql += sql[prev_position:]
                break

            new_sql += sql[prev_position: match.start()]

            variable = match.group(1)

            if variable not in ns:
                raise Exception(f"{variable} not found as Jupyter variable")
            else:
                var = ns[variable]
                rendered = SqlParser.render_type(var, variable)
                new_sql += rendered

            prev_position = match.end()

        return new_sql

    @staticmethod
    def render_type(value, variable_name):
        if isinstance(value, str):
            return value
        elif isinstance(value, pd.DataFrame):
            if variable_name is None:
                raise Exception("DataFrame type must have a name")

            value = value.copy()
            value.infer_objects()
            return SqlParser.render_dataframe(value) + \
                " as `" + variable_name + "`"
        elif isinstance(value, dict):
            if variable_name is None:
                raise Exception("DataFrame type must have a name")

            return SqlParser.render_dict(value)
        elif isinstance(value, list):
            return SqlParser.render_list(value)

    @staticmethod
    def from_datetime64_ns(value):
        if value.nanosecond != 0:
            raise Exception("No support for nanoseconds of Pandas datetime64")

        value_str = value.strftime("%Y-%m-%d %H:%M:%S.%f")
        formatted = f'DateTime::MakeTimestamp(DateTime::Parse("%Y-%m-%d %H:%M:%S")("{value_str}"))'  # noqa
        return formatted

    @staticmethod
    def from_datetime(value):
        value_str = value.strftime("%Y-%m-%d %H:%M:%S.%f")
        formatted = f'DateTime::MakeTimestamp(DateTime::Parse("%Y-%m-%d %H:%M:%S")("{value_str}"))'  # noqa
        return formatted

    @staticmethod
    def from_str(value):
        value = value.replace('"', '\\"')
        return f"\"{value}\""

    @staticmethod
    def from_int(value):
        return f"{value}l"

    @staticmethod
    def render_value(value: Any) -> str:
        if isinstance(value, str):
            value = SqlParser.from_str(value)
        elif isinstance(value, int):
            value = SqlParser.from_int(value)
        elif isinstance(value, pd._libs.tslibs.timestamps.Timestamp):
            value = SqlParser.from_datetime64_ns(value)
        elif isinstance(value, datetime):
            value = SqlParser.from_datetime(value)
        elif isinstance(value, float):
            pass
        else:
            assert not f"Unsupported  type {type(value)}"

        return value

    @staticmethod
    def render_dict(dict_value: dict) -> str:
        sql = "ToDict(AsList("

        as_dict_cols = []

        key_types = set()
        value_types = set()
        for key, value in dict_value.items():
            key_types.add(type(key))
            value_types.add(type(value))

        if len(key_types) > 1:
            raise Exception(f"All key types must be of one type. Found several {key_types}")

        if len(value_types) > 1:
            raise Exception(f"All value types must be of one type. Found several {value_types}")

        for key, value in dict_value.items():
            key = SqlParser.render_value(key)
            value = SqlParser.render_value(value)

            as_dict_cols.append(f"asTuple({key}, {value})")

        sql += ",".join(as_dict_cols)
        sql += "))"

        return sql

    @staticmethod
    def render_list(list_value: list) -> str:
        sql = "AsList("

        as_list_items = []

        for value in list_value:
            value = SqlParser.render_value(value)

            as_list_items.append(str(value))

        sql += ",".join(as_list_items)
        sql += ")"

        return sql

    @staticmethod
    def render_dataframe(df: pd.DataFrame) -> str:
        sql = "AS_TABLE(AsList("

        columns = []
        for colname, _ in df.dtypes.items():
            columns.append(colname)

        as_structs = []
        for _, row in df.iterrows():
            as_struct_cols = []
            for index, (_, value) in enumerate(row.items()):
                value = SqlParser.render_value(value)

                as_struct_cols.append(f"{value} as `{columns[index]}`")

            as_struct_row = "AsStruct(" + ",".join(as_struct_cols) + ")"
            as_structs.append(as_struct_row)

        as_structs = ",".join(as_structs)
        sql += as_structs
        sql += "))"

        return sql
